package de.erethon.hephaestus;

import de.erethon.spellbook.api.SpellLibrary;
import net.minecraft.core.registries.BuiltInRegistries;
import net.minecraft.resources.ResourceLocation;
import net.minecraft.world.item.Item;
import org.bukkit.Bukkit;
import org.bukkit.Material;
import org.bukkit.configuration.file.YamlConfiguration;
import org.bukkit.craftbukkit.v1_20_R3.inventory.CraftItemStack;
import org.bukkit.craftbukkit.v1_20_R3.util.CraftMagicNumbers;
import org.bukkit.inventory.ItemStack;
import org.bukkit.plugin.Plugin;
import org.bukkit.plugin.java.JavaPlugin;
import org.jetbrains.annotations.Nullable;

import java.io.File;
import java.io.IOException;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Set;
import java.util.logging.Level;

import static net.minecraft.server.MinecraftServer.LOGGER;

public class ItemLibrary {

    private final File folder;

    private final HashMap<ResourceLocation, HItem> items = new HashMap<>();
    private final HashMap<Material, HItem> vanillaMaterials = new HashMap<>();
    private final Set<HItem> itemsWithBehaviour = new HashSet<>();
    private HBukkitHandler bukkitHandler = null;

    public ItemLibrary(File folder) {
        this.folder = new File(folder + "/plugins/Hephaestus/items");
        if (!folder.exists()) {
            folder.mkdirs();
        }
        generateVanillaItemFiles();
        loadItems();
    }

    public boolean has(ResourceLocation key) {
        return items.containsKey(key);
    }

    @Nullable
    public HItem get(ResourceLocation key) {
        return items.get(key);
    }

    @Nullable
    public HItem get(Material material) {
        return vanillaMaterials.get(material);
    }

    @Nullable
    public HItem get(ItemStack bukkitStack) {
        CraftItemStack craftItemStack = (CraftItemStack) bukkitStack;
        if (craftItemStack.handle.hItemStack == null) {
            return null;
        }
        return craftItemStack.handle.hItemStack.config();
    }

    @Nullable
    public HItem get(net.minecraft.world.item.ItemStack stack) {
        return stack.hItemStack.config();
    }

    public boolean registerItem(Plugin plugin, HItem item) {
        if (bukkitHandler == null) { // Use the first plugin we get to register the bukkit handler, because we need a Plugin for listeners.
            bukkitHandler = new HBukkitHandler((JavaPlugin) plugin, this);
        }
        if (item.getId().getNamespace().equals("minecraft")) {
            throw new RuntimeException(plugin.getName() + " tried registering a custom item in the vanilla namespace: " + item.getId().toString());
        }
        if (items.containsKey(item.getId())) {
            Bukkit.getLogger().log(Level.WARNING, "[Hephaestus] Item with ID " + item.getId() + " is already registered.");
            return false;
        }
        item.setPlugin(plugin);
        items.put(item.getId(), item);
        return true;
    }

    public boolean overrideVanillaBehaviour(Plugin plugin, HItem item) {
        if (bukkitHandler == null) {
            bukkitHandler = new HBukkitHandler((JavaPlugin) plugin, this);
        }
        item.setPlugin(plugin);
        return true;
    }


    public boolean unregisterItem(HItem item) {
        if (!items.containsKey(item.getId())) {
            Bukkit.getLogger().log(Level.WARNING, "[Hephaestus] Item with ID " + item.getId() + " is not registered.");
            return false;
        }
        if (item.getId().getNamespace().equals("minecraft")) {
            throw new RuntimeException("Vanilla items can not be unregistered.");
        }
        items.remove(item.getId());
        return true;
    }

    private void generateVanillaItemFiles() {
        for (Item item : BuiltInRegistries.ITEM.stream().toList()) {
            File vanillaFolder = new File(folder, "vanilla");
            if (!vanillaFolder.exists()) {
                vanillaFolder.mkdirs();
            }
            YamlConfiguration configuration = new YamlConfiguration();
            ResourceLocation location = BuiltInRegistries.ITEM.getKey(item);
            configuration.set("id", location.toString());
            configuration.set("mojangID", location.toString());
            configuration.set("maxStackSize", item.getMaxStackSize());
            configuration.set("maxDurability", item.getMaxDamage());
            File file = new File(vanillaFolder, location.getPath() + ".yml");
            if (file.exists()) {
                return;
            }
            try {
                configuration.save(file);
            } catch (IOException e) {
                throw new RuntimeException(e);
            }
        }
    }

    private void loadItems() {
        for (File file : SpellLibrary.getFilesForFolder(folder)) {
            if (!file.getName().contains(".yml")) {
                continue;
            }
            HItem config = new HItem(file);
            items.put(config.getId(), config);
            vanillaMaterials.put(CraftMagicNumbers.getMaterial(config.getBaseItem()), config);
            if (config.hasBehaviour()) {
                itemsWithBehaviour.add(config);
            }
        }
        LOGGER.info("[Hephaestus] Loaded " + items.keySet().size() + " items.");
    }

    public File getFolder() {
        return folder;
    }

    public void readyBehaviours() {
        for (HItem item : itemsWithBehaviour) {
            item.loadBehaviour();
        }
    }

    public void enableHandler(Plugin plugin) {
        if (bukkitHandler == null) { // Use the first plugin we get to register the bukkit handler, because we need a Plugin for listeners.
            bukkitHandler = new HBukkitHandler((JavaPlugin) plugin, this);
        }
    }

    public static @Nullable HItemStack getHItemFromBukkit(ItemStack stack) {
        CraftItemStack craftItemStack = CraftItemStack.asCraftCopy(stack);
        if (craftItemStack.handle == null) { // why in the fuck are there craft stacks that have no nms stack???
            return null;
        }
        return craftItemStack.handle.hItemStack;
     }
}
